#ifndef __CS_H__
#define __CS_H__

#include "NGM.h"

#include "SequenceLocation.h"
#include "MappedRead.h"
#include "RefEntry.h"

#include <map>
#include <stdlib.h>
#include <string>

#include "CSCache.h"

/*
Candidate Search:

Initialisierung:
  Iteriere ueber jeden Praefix der definierten Referenzen (config:ref_mode mit -1=alle, 0..n eine einzelne referenz)
    Fuege in RefTable fuer diesen Praefix seine Position und die Id des Referenzstrings ein

Suche:
  Iteriere ueber jeden Pruefix des Reads
    Lies fuer diesen Praefix die Positionsliste aus RefTable
	Fuer jede Position dieses Praefixes
	  Offset um die Position des Praefixes (=normalisierung auf den Read-Anfang)
	  fast_mode:
	    Pruefe rTable-Capactiy, wenn overflow -> Neustart der Suche im safe_mode
	    Schreibe in rTable mit linearem Kollisionshandling
	  save_mode:
	    Schreibe in iTable

  Sammle Ergebnisse ein:
    Iteriere ueber jedes Element in fast_mode:rTable / save_mode:iTable
	  Wenn count > threshhold, schicke es zur weiterverarbeitung

*/
class CS : public NGMTask, private virtual IRefProvider
{
private:
	friend class CSCache;

	volatile static int s_ThreadCount;
	int const m_CSThreadID;
	int m_CurrentStart;
	int m_CurrentCount;
	int m_CurrentSeq;
	int m_CurrentPage;
	int m_WrittenReads;
	int m_DiscardedReads;
	volatile int m_Candidates;
	int m_Mode;
	MappedRead * * m_ReadBatch;

	IRefProvider const * m_RefProvider;

	static int cSrchTableLen;
	static int cThreshhold;
	static uint prefixBasecount;
	static uint prefixBits;
	static ulong prefixMask;
	static uint prefixBaseSkip;

	static bool useCache;
	static std::string cacheFileName;

	typedef void (*PrefixIterationFn)(ulong prefix, int pos, void* data);

	static void PrefixIteration(char const * sequence, uint length, PrefixIterationFn func, void* data, uint offset = 0);
	static void BuildPrefixTable(ulong prefix, int pos, void* data);
	static void PrefixSearch(ulong prefix, int pos, void* data);
	int CollectResults(MappedRead * read, int mode, int refnr);
	void FilterScore(LocationScore * score);

	void RunBatch();
	void SendToBuffer(MappedRead * read);

	//---section IRefProvider
	int GetPageCount() const
	{
		return 1;
	}
	RefEntry * GetRefEntry(ulong prefix) const;
	void LoadPage(int const page) const
	{
		return;
	}
	//---section IRefProvider

	static void SaveToRefTable(ulong prefix, SequenceLocation loc);
	static void PackRefTable();
	static void ClearRefTable();

	void CheckRTable(); // Debug/Validation

	LocationScore * rTable;	// standard
	std::map<SequenceLocation, int> iTable;		// fallback

	static inline int hash(int n) { return abs(n) % cSrchTableLen; }

	static inline int calc_binshift(int corridor)
	{
		corridor >>= 1;
		int l = 0;
		while ( (corridor >>= 1) > 0 )
			++l;
		return l;
	}

	inline int GetBin(int pos)
	{
		static int shift = calc_binshift(Config.GetInt("corridor"));
		return pos >> shift;
	}
	inline int ResolveBin(int bin)
	{
		static int shift = calc_binshift(Config.GetInt("corridor"));
		static int offset = (shift > 0) ? 1 << (shift - 1) : 0;
		return (bin << shift) + offset;
	}
public:
	static int cRefTableLen;

	//inline const int GetCurrentQrySeq() { return m_CurrentStart + i; }
	inline const int GetMode() { return m_Mode; }

	static void Init();
	static void Cleanup();

	CS(bool useBuffer = true) : m_CSThreadID( (useBuffer) ? (AtomicInc( &s_ThreadCount ) - 1) : -1 ),
			m_WrittenReads( 0 ),
			m_DiscardedReads( 0 )
	{
		if (useBuffer)
			NGM.bCSSW.Register();
		rTable = 0;
	}
	~CS();

	void DoRun();
	inline int GetStage() const { return 0; }
	inline char const * GetName() const { return "CS"; }
};

static inline int CSHash(ulong prefix)
{
	return prefix % CS::cRefTableLen;
}

const int cInvalidLocation = -99999999;
const SequenceLocation sInvalidLocation = { cInvalidLocation, 0 };
const LocationScore sInvalidLocationScore = { sInvalidLocation, {0} };

#endif
