#include <stdio.h>
#include <memory.h>

#include "RefEntry.h"
#include "Types.h"

#include "Log.h"
#include "NGM.h"

#include "CS.h"

#include "CSCache.h"

#undef module_name
#define module_name "CSCache"

const uint cMaxCachePages = 62;

struct CSCachePageIndex
{
	uint Data;
	uint RefTableLength;
	uint CheckSum;
	uint PageCount;
	ulong Reserved;
	ulong MaxPageSize;	// To allocate memory when loading a page
	ulong Offsets[cMaxCachePages];
	ulong Lengths[cMaxCachePages];
};

ulong const cPageSig = 0xF12233445566778F;

uint const cSigMore = 0xF11FF11F;
uint const cSigLast = 0xF00FF00F;

struct Entry
{
	ulong Prefix;
	uint Count;
	uint Next;
};

CSCache * CSCache::CreateCache(std::string const filename)
{
	CSCache * cache = new CSCache();
	cache->m_Filename = filename;
	cache->m_FP = fopen(filename.c_str(), "w+b");

	if (cache->m_FP == 0)
	{
		Log.Error("Unable to create page cache file.");
		Fatal();
	}

	cache->m_Index = new CSCachePageIndex();
	size_t const size = sizeof(CSCachePageIndex);
	memset(cache->m_Index, 0, size);

	cache->m_Index->Data = CS::prefixBasecount;
	cache->m_Index->RefTableLength = CS::cRefTableLen;

	if (fwrite(cache->m_Index, size, 1, cache->m_FP) != 1)
	{
		Log.Error("Error writing cache file header.");
		Fatal();
	}

	return cache;
}

CSCache const * CSCache::LoadCache(std::string const filename)
{
	CSCache const * cache = new CSCache();
	cache->m_Filename = filename;
	cache->m_FP = fopen(filename.c_str(), "rb");

	if (cache->m_FP == 0)
	{
		Log.Error("Unable to open page cache file.");
		return 0;
	}

	size_t const size = sizeof(CSCachePageIndex);
	char * buf = new char[size];

	size_t read = fread(buf, size, 1, cache->m_FP);
	if (read != 1)
	{
		Log.Error("Error reading cache file header");
		return 0;
	}

	cache->m_Index = (CSCachePageIndex*)buf;

	if (cache->m_Index->Data != CS::prefixBasecount)
	{
		Log.Error("Mismatching prefix size in cache (Cache: %i Bases, Config: %i Bases)", cache->m_Index->Data, CS::prefixBasecount);
		Fatal();
	}

	return cache;
}

int WritePrefixRefs(PrefixRefRelay * relay, FILE * fp)
{
	RefEntry * cur = relay->Entry;

	int total = relay->Entry->refTotal;
	int written = 0;

	Entry entry;
	entry.Prefix = relay->Prefix;
	entry.Count = total;
	entry.Next = (relay->Next == 0) ? cSigLast : cSigMore;

	fwrite(&entry, sizeof(Entry), 1, fp);

	while (cur != 0)
	{
		written = fwrite(cur->ref, sizeof(SequenceLocation), cur->refCount, fp);
		total -= written;

		if (written != cur->refCount)
		{
			Log.Error("Error writing data to cache file (%i of %i written)", written, cur->refCount);
			Fatal();
		}

		cur = cur->nextEntry;
	}

	if (total != 0)
	{
		Log.Error("Count mismatch (%i) writing prefix 0x%x to cache file", total, relay->Prefix);
		Fatal();
	}

	return sizeof(Entry) + sizeof(SequenceLocation) * relay->Entry->refTotal;
}

void CSCache::AddPage(PrefixRefRelay * * Table, uint const length)
{
	if (m_Index->PageCount >= cMaxCachePages)
	{
		Log.Error("Cache file %s cant hold any more pages (maximum of %i reached)", m_Filename.c_str(), cMaxCachePages);
		Fatal();
	}

	if (length != m_Index->RefTableLength)
	{
		Log.Error("Mismatching ref table length in page file for adding page (file = %i, added page = %i)", m_Index->RefTableLength, length);
		Fatal();
	}

	Log.Message("Adding cache page %i", m_Index->PageCount);

	ulong totalPageSize = 0;
	fseek(m_FP, 0, SEEK_END);	// Neue Seite am Dateiende hinzufügen

	ulong pageStart = ftell(m_FP); // Beginn der neuen Seite

	fwrite(&cPageSig, sizeof(ulong), 1, m_FP);
	totalPageSize += sizeof(ulong);

	ulong tableSize = (length+1)*sizeof(ulong); // Größe der Hashtabelle
	if (fseek(m_FP, tableSize, SEEK_CUR) != 0) // Skippen
	{
		Log.Error("Error writing page index data");
		Fatal();
	}
	totalPageSize += tableSize;

	ulong * offsetTable = new ulong[length+1]; // Temporäre Hash/Offsettabelle
	memset(offsetTable, 0, tableSize);
	offsetTable[length] = 0x11AAAAAAAAAAAA11;
	NGM.Stats->CachePageWriteMax = length;

	for (uint i = 0; i < length; ++i)
	{
		NGM.Stats->CachePageWriteCur = i;
		PrefixRefRelay * cur = Table[i];
		if (cur != 0)
			offsetTable[i] = ftell(m_FP) - pageStart;

		while (cur != 0)
		{
			totalPageSize += WritePrefixRefs(cur, m_FP);
			cur = cur->Next;
		}
	}

	fseek(m_FP, pageStart+sizeof(ulong), SEEK_SET);
	fwrite(offsetTable, tableSize, 1, m_FP);

	if (totalPageSize > m_Index->MaxPageSize)
	{
		if (m_pMem != 0)
		{
			delete m_pMem;
			m_pMem = 0;
		}

		m_Index->MaxPageSize = totalPageSize;
	}

	m_Index->Offsets[m_Index->PageCount] = pageStart;
	m_Index->Lengths[m_Index->PageCount] = totalPageSize;
	++m_Index->PageCount;

	rewind(m_FP);
	fwrite(m_Index, sizeof(CSCachePageIndex), 1, m_FP);
	fflush(m_FP);
}

void CSCache::LoadPage(int page) const
{
	if (page >= GetPageCount() || page < 0)
	{
		Log.Error("Page number %i out of range", page);
		return;
	}
	if (m_Index->Offsets[page] == 0 || m_Index->Lengths[page] == 0)
	{
		Log.Error("No data for page %i. This could indicate a corrupt cache file.", page);
		Fatal();
	}

	if (m_CurrentPage == page)
		return;

	m_CurrentPage = page;

	if (m_pMem == 0)
	{
		m_pMem = new char[m_Index->MaxPageSize];
	}

	fseek(m_FP, m_Index->Offsets[m_CurrentPage], SEEK_SET);
	if (fread(m_pMem, m_Index->Lengths[m_CurrentPage], 1, m_FP) != 1)
	{
		Log.Error("Error loading page %i", m_CurrentPage);
	}

	if ( *((ulong*)m_pMem) != cPageSig )
	{
		Log.Error("Invalid page signature");
		Fatal();
	}
}

int  CSCache::GetPageCount() const
{
	return m_Index->PageCount;
}

RefEntry * CSCache::GetRefEntry(ulong prefix) const
{
	if (m_CurrentPage == -1)
		return 0;

	int hash = CSHash(prefix);
	ulong * table = (ulong*)(m_pMem + sizeof(ulong));
	ulong offset = table[hash];
	if (offset == 0)
		return 0;

	char * ptr = m_pMem + offset;
	Entry * e = (Entry*)ptr;

	while (e->Prefix != prefix && e->Next == cSigMore)
	{
		e = (Entry*)(ptr + sizeof(Entry) + sizeof(SequenceLocation) * e->Count);
	}

	if (e->Next != cSigLast && e->Next != cSigMore)
	{
		Log.Error("Invalid entry signature for prefix 0x%x (hash 0x%x) on page %i (Offset = 0x%x, Next = 0x%x)", prefix, hash, m_CurrentPage, offset, e->Next);
		Fatal();
	}

	if (e->Prefix == prefix)
	{
		return new RefEntry((SequenceLocation*)(e + 1), e->Count);
	}
	else
		return 0;
}

CSCache::CSCache() :
		m_pMem( 0 ),
		m_Index( 0 ),
		m_CurrentPage( -1 )
{
}
CSCache::~CSCache()
{
	fclose(m_FP);
	if (m_pMem)
		delete[] m_pMem;
	if (m_Index != 0)
		delete[] m_Index;
	m_pMem = 0;
	m_CurrentPage = -1;
}
