#include "Config.h"

#include "PlatformSpecifics.h"
#include "Log.h"
#include "NGM.h"

#include <memory.h>
#include <stdlib.h>
#include <stdio.h>

#include <algorithm>

#include "Debug.h"

#undef module_name
#define module_name "CONFIG"

std::string add_timestamp(std::string);	// from logging.cpp

	// Config is stored as a string to string map, with
	// values converted just in time for the user to
	// provide type information rather then trying to find
	// out ourselves

bool _Config::InternalExists(std::string name) const
{
	std::transform(name.begin(), name.end(), name.begin(), ::tolower);
	return config_map->count(name) == 1;
}

std::string _Config::InternalGet(std::string name, char const * * arr_data) const
{
	std::transform(name.begin(), name.end(), name.begin(), ::tolower);
	if (!InternalExists(name))
	{
		Log.Warning("Tried to access unknown config value \"%s\"", name.c_str());
		return "";
	}
	else
	{
		if (arr_data != 0)
		{
			if (config_arrays->count(name) == 0)
			{
				Log.Error("No array declared for %s", name.c_str());
			}
			else
			{
				*arr_data = (*config_arrays)[name].c_str();
			}
		}
		return (*config_map)[name];
	}
}
std::string _Config::InternalGet(std::string name) const
{
	return InternalGet(name, 0);
}

void _Config::InternalAdd(std::string name, std::string value, std::string arr_data = std::string(), bool override)
{
	std::transform(name.begin(), name.end(), name.begin(), ::tolower);
	bool exists = InternalExists(name);
	if (exists && !override)
	{
		Log.Warning("Ignoring redefinition of %s (current value = %s, tried to override with %s)", name.c_str(), InternalGet(name).c_str(), value.c_str());
	}
	else
	{
		if (name == "logfile" || name == "output")
			value = add_timestamp(value);

		if (name == "qry_end")
			Log.Warning("Parameter qry_end has been replaced by qry_count");

		if(name != "cmdline") {
			Log.Message("%s <%s : %s>%s", (exists) ? "Overriding" : "Adding", name.c_str(), value.c_str(), (arr_data.length() != 0) ? "[Array]" : "");
		}
		config_map->insert(std::pair<std::string, std::string>(name, value));
		if (arr_data.length() != 0)
			config_arrays->insert(std::pair<std::string, std::string>(name, arr_data));
	}
}

extern char const * const defaultConfigText;

void CreateDefaultConfig()
{
	Log.Message("Writing default config");

	FILE * f = fopen("config.default", "w");
	if (f == 0)
	{
		Log.Error("Unable to create file");
		Fatal();
	}
	fprintf(f, "%s", defaultConfigText);
	fclose(f);
	exit(0);
}

void ParseArguments(int argc, char * argv[]);

/*_Config const & _Config::Instance()
{
	if (pInstance == 0)
	{
		pInstance = new _Config();
	}

	return *pInstance;
}*/

char const * _Config::GetString(char const * const name) const
{
	return InternalGet(name).c_str();
}

int GetConversionFactor(std::string s)
{
	char c = s[s.length()-1];
	if (c == 'k')
		return 1000;
	else if (c == 'm')
		return 1000*1000;
	else if (c == 'g')
		return 1000*1000*1000;
	else
		return 1;
}

int _Config::GetInt(char const * const name) const
{
	std::string const str = InternalGet(name);
	if (!str.empty())
		return atoi(str.c_str())*GetConversionFactor(str);
	else
		return 0;
}

int _Config::GetParameter(char const * const name) const
{
	return GetInt(name);
}

float _Config::GetFloat(char const * const name) const
{
	std::string const str = InternalGet(name);
	if (!str.empty())
		return (float)atof(str.c_str()) * (float)GetConversionFactor(str);
	else
		return 0.0f;
}

int _Config::GetInt(char const * const name, int min, int max) const
{
	int i = GetInt(name);
	if ( (i < min) || ((max >= min) && (i > max)) )
	{
		Log.Error("Value %s : %i out of range [%i, %i]", name, i, min, max);
		Fatal();
	}
	return i;
}
float _Config::GetFloat(char const * const name, float min, float max) const
{
	float f = GetFloat(name);
	if ( (f < min) || ((max >= min) && (f > max)) )
	{
		Log.Error("Value %s : %f out of range [%f, %f]", name, f, min, max);
		Fatal();
	}
	return f;
}

void _Config::Override(char const * const name, char const * const value)
{
	InternalAdd(name, value, "", true);
}
void _Config::Override(char const * const name, int const value)
{
	char buffer[32];
	sprintf(buffer, "%i", value);
	Override(name, buffer);
}

// Helper functions for config file parsing
inline bool IsWhitespace(char const * const str)
{
	return (*str == ' ' || *str == '\n' || *str == '\t' || *str == '\r');
}
inline void SkipNonWhitespace(char const * & str)
{
	while (*str != 0 && !IsWhitespace(str))
		++str;
}
inline void SkipWhitespace(char const * & str)
{
	while ( *str != 0 && IsWhitespace(str) )
		++str;
}
inline void SkipLine(char const * & str, bool stopForArray)
{
	while (*str != 0 && *str != '\n' && *str != '\r' && !(stopForArray && *str == '{'))
	{
		++str;
	}
	if (*str != '{')
		++str;
}

// Converts the stored matrix into actual numbers by the given conversion function
// and writes them into the data array
template <typename T> int ParseMatrix(_Config const * config, char const * const name, T * data, int len, T f(char const * const))
{
	char const * matrix_str = 0;

	char const * arr_size_str = config->InternalGet(name, &matrix_str).c_str();
	if (arr_size_str == 0)
	{
		Log.Error("Invalid matrix \"%s\"", name);
		Fatal();
	}

	int internal_len = atoi(arr_size_str);

	if (matrix_str == 0)
		return 0;

	if (len != internal_len)
		Log.Warning("Requested matrix size doesn't match declaration in config file");

	int i = 0;
	for (; i < len; ++i)
	{
		SkipWhitespace(matrix_str);
		data[i] = f(matrix_str);
		SkipNonWhitespace(matrix_str);
	}
	return i;
}

int _Config::GetIntArray(char const * const name, int * pData, int const len) const
{
	return ParseMatrix<int>(this, name, pData, len, atoi);
}

// strangly enough, the c standard declares atof as returning double
float _atof(char const * const str)
{
	return (float)atof(str);
}

int _Config::GetFloatArray(char const * const name, float * pData, int const len) const
{
	return ParseMatrix<float>(this, name, pData, len, _atof);
}

int _Config::GetDoubleArray(char const * const name, double * pData, int len) const
{
	return ParseMatrix<double>(this, name, pData, len, atof);
}

bool _Config::Exists(char const * const name) const
{
	return InternalExists(name);
}

bool _Config::HasArray(char const * const name) const
{
	return config_arrays->find(std::string(name)) != config_arrays->end();
}

// Simple config file parser
void _Config::ParseFile(char const * const filename)
{
	if (sealed)
	{
		Log.Error("Cant parse config file - config has been sealed.");
		return;
	}

	if (!FileExists(filename))
	{
		Log.Error("Couldnt find config file \"%s\".", filename);
		return;
	}

	char const * config_text = 0;
	int map_id = CreateMapping(filename, config_text);

	while (*config_text != 0)
	{
		if (*config_text == '#')
			SkipLine(config_text, false);

		SkipWhitespace(config_text);

		if (*config_text != 0 && *config_text != '#')
		{
			int n = 0;

			while (*(config_text+n) != 0 && *(config_text+n) != ' ')
				++n;

			std::string name = std::string(config_text, n);
			config_text += n;

			SkipWhitespace(config_text);

			if (*config_text == 0 || *config_text != '=')
				Log.Error("'=' expected after identifier '%s'", name.c_str());
			else
			{
				++config_text;

				SkipWhitespace(config_text);

				char const * start = config_text;

				SkipNonWhitespace(config_text);

				ulong len = (ulong)config_text - (ulong)start;
				std::string value(start, len);
				
				SkipLine(config_text, true);
				std::string arr_str;
				if (*config_text == '{')
				{
					++config_text;
					start = config_text;

					while (*config_text != '\0' && *config_text != '}')
						++config_text;

					if (*config_text != '}')
						Log.Error("'}' expected.");
					else
					{
						++config_text;
					}
					arr_str = std::string(start, config_text-start);
				}

				InternalAdd(name, value, arr_str);
			}
		}
	}

	CloseMapping(map_id);
}

// Initializes config by parsing arguments from command line
// and a config file
_Config::_Config(int argc, char * argv[])
{
	initialized = false;
	sealed = false;

	config_map = new TConfigMap();
	config_arrays = new TConfigMap();

	ParseArguments(argc, argv);

	if (InternalExists("config"))
	{
		ParseFile(InternalGet("config").c_str());
	}

	initialized = true;

	//Add full command line for CIGARWriter
	std::string cmdLine;
	for(TConfigMap::iterator it = config_map->begin(); it != config_map->end(); ++it) {
	  cmdLine += " -c:" + it->first + " " + it->second;
	}
	InternalAdd("cmdline", cmdLine, "", true);
}

//IConfig::~IConfig() {}

_Config::~_Config()
{
	delete config_arrays;
	delete config_map;
}

void _Config::ParseArguments(int argc, char * argv[])
{
	if (argc == 1)
	{
		Log.Error("Missing arguments. Usage:\n");
		Help();
	}
	// one parameter only
	if (argc == 2)
	{
		// check for help "-h..." or default config creation "-d..."
		if (argv[1][0] == '-')
		{
			if (argv[1][1] == 'h')
				Help();
			else if(argv[1][1] == 'd')
				CreateDefaultConfig(); // Create Default Config
		}
		else // otherwise treat it as config file name
		{
			InternalAdd("config", argv[1]);
		}
	}
	else
	{
		for (int i = 1; i < argc; ++i)
		{
			if (argv[i][0] == '-')
			{
				// check for at least one more argument after the current,
				// except for the -h (help) switch
				if ((argc > i+1) || (argv[i][1] == 'h'))
				{
					switch (argv[i][1])
					{
					case 'c':
						if (argv[i][2] == ':')
							InternalAdd(argv[i]+3, argv[i+1]);	// Directly add config value via console ("-c:name value")
						else
							InternalAdd("config", argv[i+1]);	// Config file ("-c configfile")
						++i;
						break;
					case 'r':
						InternalAdd("ref", argv[++i]);			// reference file
						break;
					case 'i':
						InternalAdd("qry", argv[++i]);			// qry file
						break;
					case 'o':
						InternalAdd("output", argv[++i]);			// output file
						break;
					case 'p':
						InternalAdd("config", argv[++i]);			// alternate method for setting config file
						break;
					case 'l':
						InternalAdd("logfile", argv[++i]);
						InternalAdd("log", "1");
						break;
					case 'h':
						Help();
						break;
					case 'd':
						CreateDefaultConfig();
						break;
					default:
						Log.Warning("Unknown parameter %s", argv[i]);
						break;
					}
				}
				else
				{
					Log.Error("Missing value for parameter %s", argv[i]);
				}
			}
			else
			{
				Log.Warning("Unknown parameter \"%s\"", argv[i]);
			}
		}
	}
}

char const * const defaultConfigText =
"################################################################################\n\
#                                                                              #\n\
# NGM Config Example                                                           #\n\
#                                                                              # \n\
# Lines starting with \"#\" are comments and thus ignored                        #\n\
#                                                                              #\n\
################################################################################\n\
# General                                                                      #\n\
################################################################################\n\
ref = ../data/refseq\n\
qry = ../data/reads\n\
pqry = ../data/bqryfile.dat	# 2nd query file for paired reads\n\
output = results-%s.txt		# add %s to insert a data/timestamp\n\
\n\
################################################################################\n\
# Input Data Conversion                                                        #\n\
################################################################################\n\
use_cache = 1		# try using the cached files\n\
ref_type = 0		# 0 .. FASTA, 1 .. Readfile\n\
qry_type = 1\n\
convert_verbose = 1	# display output from conversion tool?\n\
\n\
# Sequence\n\
qry_max_len = 74\n\
paired = 0		# paired read mode -> use pqry to specify input\n\
dualstrand = 1		# set to 1 to enable search on + and - strand\n\
\n\
################################################################################\n\
# Input Data Selection                                                         #\n\
################################################################################\n\
ref_mode = 0		# -1 = all refs, 0...n = single ref\n\
\n\
qry_start = 0		# start at this read\n\
qry_end = 20		# up to this read. set to 0 to iterate over all reads\n\
\n\
################################################################################\n\
# Execution config                                                             #\n\
################################################################################\n\
cs_threads = 2\n\
cs_batchsize = 100000	# default: 1000\n\
\n\
sw_threads = 1 { 0 }	# sw configuration: the array defines the gpus to use\n\
                        # in each thread\n\
\n\
################################################################################\n\
# Logging                                                                      #\n\
################################################################################\n\
log = 0				# use logfile param to specify a filename\n\
				# if no other filename is given, \"./ngm_%s.log\"\n\
                                # is used\n\
logfile = ./ngm_%s.log		# use %s to insert a timestamp into the filename\n\
stderr = 1	# not yet implemented: print errors to stderr rather then stdout\n\
\n\
################################################################################\n\
# Candidate search                                                             #\n\
################################################################################\n\
cs_prefixsize = 10		# Prefix length in bases for CS\n\
cs_threshhold = 4		# Amount of Prefixes that have to be found at a\n\
                                # specific location\n\
\n\
cs_maxRefsPerEntry = 100	# Allocation chunk size for the reference table\n\
cs_searchTableLen = 23333       # Hashtable size for candidate search\n\
# Should be a prime number. Set to 0 to always use dynamic allocation (slower).\n\
# Higher numbers mean less collisions/overflows, gaining performance\n\
# Too high values will degrade performance\n\
\n\
################################################################################\n\
# Alignment                                                                    #\n\
################################################################################\n\
score_match = 4\n\
score_mismatch = -2\n\
score_gap_read = -7\n\
score_gap_ref = -7\n\
sw_mode = 0		# 0..local alignment\n\
\n\
\n\
################################################################################\n\
# Arrays can be defined with \"n {...}\", n being the number of elements\n\
# in the array. The opening bracked HAS to be on the same line as the \n\
# defining name\n\
example_array_or_matrix = 16 {\n\
 1  2  3  4\n\
 5  6  7  8\n\
 9 10 11 12\n\
13 14 15 22\n\
}\n";
