#include <stdio.h>

#include "Output.h"
#include "Debug.h"

#include <string.h>

#include "FFormatWriter.h"
#include "CIGARWriter.h"

static inline void rev(char * s);

char const * const cFormatNames[] = { "Plain Text", "SAM" };

void Output::DoRun()
{
	char const * const output_name = Config.GetString("output");
	int const outputformat = NGM.GetOutputFormat();

	Log.Message("Writing output to %s (Format: %s)", output_name, cFormatNames[outputformat]);

	m_Writer = (outputformat == 0) ? (GenericReadWriter*)new FFormatWriter(output_name) : (GenericReadWriter*)new CIGARWriter(output_name);
	m_Writer->WriteProlog();

	int const batchSize = NGM.Aligner()->GetAlignBatchSize();
	Log.Message("Alignment batchsize = %i", batchSize);

	int const corridor = Config.GetInt("corridor");
	
	MappedRead ** reads = new MappedRead*[batchSize];

	char const * * qryBuffer = new char const *[batchSize];
	char const * * refBuffer = new char const *[batchSize];
	Align * alignBuffer = new Align[batchSize];
		
	while (NGM.StageActive(GetStage() - 1) || NGM.bSWO.Count() > 0)
	{
		int count = NGM.bSWO.Read(reads, batchSize);

		if (count > 0)
		{
			for (int i = 0; i < count; ++i)
			{
				MappedRead * cur_read = reads[i];
				cur_read->Strand = Strand(cur_read);

				// Bei Mapping am Minus-Strang, Position bez�glich +-Strang reporten
				if (cur_read->Strand == '-')
				{

					int read_len = SequenceProvider.GetQryLen(cur_read->ReadId);
					int max_len = SequenceProvider.GetReadMaxLen();

					cur_read->Seq = new char[max_len];
					memcpy(cur_read->Seq, SequenceProvider.GetQrySequence(cur_read->ReadId), max_len);
					rev(cur_read->Seq);

					qryBuffer[i] = cur_read->Seq;

					// Position am +-Strang
					cur_read->TLS()->Location.m_Location = SequenceProvider.GetRefLen(cur_read->TLS()->Location.m_RefId) -
						cur_read->TLS()->Location.m_Location - read_len;
					// RefId auf +-Strang setzen
					--cur_read->TLS()->Location.m_RefId;

					refBuffer[i] = SequenceProvider.GetRefSequence(cur_read->TLS()->Location.m_RefId) + cur_read->TLS()->Location.m_Location - (corridor >> 1);
				}
				else
				{
					qryBuffer[i] = SequenceProvider.GetQrySequence(cur_read->ReadId);
					refBuffer[i] = SequenceProvider.GetRefSequence(cur_read->TLS()->Location.m_RefId) + cur_read->TLS()->Location.m_Location - (corridor >> 1);
				}

				alignBuffer[i].pBuffer1 = cur_read->Buffer1;
				alignBuffer[i].pBuffer2 = cur_read->Buffer2;
			}

			Log.Message("Thread %i invoking alignment (count = %i)", m_TID, count);

			int aligned = NGM.Aligner()->BatchAlign(0 | (outputformat << 8), count, refBuffer, qryBuffer, alignBuffer, 0);

			if (aligned == count)
			{
				Log.Message("Align done");
			}
			else
			{
				Log.Error("Error aligning outputs (%i of %i aligned)", aligned, count);
			}

			for (int i = 0; i < aligned; ++i)
			{
				MappedRead * cur_read = reads[i];

				if (cur_read->Strand == '-')
				{
					delete cur_read->Seq;
					cur_read->Seq = 0;
				}

				cur_read->TLS()->Location.m_Location += alignBuffer[i].PositionOffset - (corridor >> 1);
				//cur_read->TLS()->Score.f = alignBuffer[i].Score; // TODO: Align liefert keine Scores
				cur_read->Identity = alignBuffer[i].Identity;
				cur_read->QStart = alignBuffer[i].QStart;
				cur_read->QEnd = alignBuffer[i].QEnd;

				if ( cur_read->EqualScoringCount > 1 )
				{
					m_EqualScoringBuffer[cur_read->ReadId].push_back(cur_read);

					if (m_EqualScoringBuffer[cur_read->ReadId].size() == cur_read->EqualScoringCount)
					{
						typedef std::list<MappedRead*> TReadList;
						TReadList reads = m_EqualScoringBuffer[cur_read->ReadId];
						reads.sort(EqualScoringSortPredicate);
						//reads.unique(EqualScoringEquivalentPredicate);
						int esc = reads.size();
						TReadList::iterator itr = reads.begin();
						TReadList::iterator last = itr;
						++itr;
						while (itr != reads.end())
						{
							if ((*last != 0) && EqualScoringEquivalentPredicate(*last, *itr))
							{
								delete *last;
								*last = 0;
								--esc;
							}
							++itr;
						}
						int esid = 0;
						for (TReadList::iterator itr = reads.begin(); itr != reads.end(); ++itr)
						{
							if (*itr != 0)
							{
								(**itr).EqualScoringID = esid++;
								(**itr).EqualScoringCount = esc;
								SaveRead(*itr);
								delete *itr;
								*itr = 0;
							}
						}
						//std::for_each(reads.begin(), reads.end(), std::bind1st( std::mem_fun(&Output::SaveRead), this ) );
					}
				}
				else
				{
					SaveRead(cur_read);
					delete cur_read;
				}
			} // for
			Log.Message("Thread %i finished batch", m_TID);
		} // if count > 0
		else
		{
			Log.Message("Nothing to do...waiting");
			Sleep(500);
		}
	} // while

	Log.Message("Freeing resources...");
	delete[] qryBuffer;
	delete[] refBuffer;
	delete[] alignBuffer;

	delete[] reads;

	Log.Message("Finalizing output...");
	m_Writer->WriteEpilog();

	delete m_Writer;

	Log.Message("Output finished");
}

static inline char cpl(char c)
{
	if (c == 'A')
		return 'T';
	else if (c == 'T')
		return 'A';
	else if (c == 'C')
		return 'G';
	else if (c == 'G')
		return 'C';
	else
		return c;
}

// swaps two bases and complements them
static inline void rc(char & c1, char & c2)
{
	char x = c1;
	c1 = cpl(c2);
	c2 = cpl(x);
}

// Reverse-complements a 0-terminated string in place
static inline void rev(char * s)
{
	char * end = s + strlen(s) - 1;

	while (s < end)
		rc(*s++, *end--);

	if (s == end)
		*s = cpl(*s);
}

char Strand(MappedRead * read)
{
	static bool sDualStrand = NGM.DualStrand();
	return (sDualStrand && (read->TLS()->Location.m_RefId & 1)) ? '-' : '+';
}
