#include "SW.h"

#include "Debug.h"

void SW::DoRun()
{
	int const corridor = Config.GetInt("corridor");

	int batchSize = NGM.Aligner()->GetScoreBatchSize();

	m_QryBuffer = new char const *[batchSize];
	m_RefBuffer = new char const *[batchSize];
	m_ScoreBuffer = new float[batchSize];
	
	LocationScore * * scores = new LocationScore*[batchSize];

	while (NGM.StageActive(GetStage()-1) || NGM.bCSSW.Count() > 0)
	{
			bool resume = true;
			if (!m_ReadBuffer.empty())
				resume = CommitReads();

			if (resume)
			{
				int count = NGM.bCSSW.Read(scores, batchSize);

				if (count > batchSize)
				{
					Log.Error("returned count=%i > batchSize=%i", count, batchSize);
					Fatal();
				}

				if (count != 0)
				{
					for (int i = 0; i < count; ++i)
					{
						m_QryBuffer[i] = SequenceProvider.GetQrySequence(scores[i]->Read->ReadId);
						m_RefBuffer[i] = SequenceProvider.GetRefSequence(scores[i]->Location.m_RefId) + scores[i]->Location.m_Location - (corridor >> 1);
						m_ScoreBuffer[i] = -1;
					}

					Log.Message("SW Thread %i launching batch (size=%i)", m_TID, count);

					int res = NGM.Aligner()->BatchScore(m_AlignMode, count, m_RefBuffer, m_QryBuffer, m_ScoreBuffer, 0);
					if (res != count)
						Log.Error("SW Kernel couldnt calculate all scores (%i out of %i)", res, count);

					Log.Message("SW Thread %i finished batch", m_TID);

					for (int i = 0; i < count; ++i)
					{
						scores[i]->Score.f = m_ScoreBuffer[i];
						if ( AtomicInc( &(scores[i]->Read->Calculated) ) == scores[i]->Read->nScores())
						{
							SendToPostprocessing(scores[i]->Read);
						}
					}
				}
				else
				{
					Log.Message("Nothing to do...waiting");
					Sleep(500);
				} // if (count...
			} // if (resume)
	} // while

	CommitReads();

	delete[] scores; scores = 0;
	delete[] m_ScoreBuffer; m_ScoreBuffer = 0;
	delete[] m_RefBuffer; m_RefBuffer = 0;
	delete[] m_QryBuffer; m_QryBuffer = 0;

	NGM.bSWO.Release();
	FinishStage();

	Log.Message("SW Thread %i finished", m_TID);
}

void SW::SendToPostprocessing(MappedRead * read)
{
	// Program runs in Paired mode and current read got a pair
	if (NGM.Paired() && (read->Paired != 0))
	{
		// Paired read finished score calculation
		// -> race condition here: both paired reads could enter at the same time
		if (read->Paired->Calculated == read->Paired->nScores())
		{
			if (read->ReadId < read->Paired->ReadId)
			{
				if (AtomicInc(&read->Lock) == 1)
				{
					PairedReadSelection(read, read->Paired);
				}
			}
			else
			{
				if (AtomicInc(&read->Paired->Lock) == 1)
				{
					PairedReadSelection(read->Paired, read);
				}
			}
		}
	}
	else
	{
		float score_max = -1;
		int score_max_loc = 0;
		int score_max_count = 0;

		// zähle anzahl topscores
		for (int j = 0; j < read->nScores(); ++j)
		{
			//Log.Verbose("Candidate %i: Loc %i, Score %f", j, read->Scores[j].Location.m_Location, read->Scores[j].Score.f);
			if (read->Scores[j]->Score.f > score_max)
			{
				score_max = read->Scores[j]->Score.f;
				score_max_loc = j;
				score_max_count = 1;
			}
			else if (read->Scores[j]->Score.f == score_max)
				++score_max_count;
		}

		static int const eslimit = Config.GetInt("EqualScoringLimit");
		if (eslimit > 0 && score_max_count > eslimit)
		{
			Log.Message("Read %i rejected for hitting equal scoring limit (%i top scores of %i)", read->ReadId, score_max_count, score_max);
			delete read;
			return;
		}

		read->TopScore = score_max_loc;
		//NGM.bSWO.Write(&read, 1);
		m_ReadBuffer.push_back(read);
		Log.Verbose("Read %i: Loc %i (Ref %i), Score %f", read->ReadId, read->TLS()->Location.m_Location, read->TLS()->Location.m_RefId, read->TLS()->Score.f);

		if (score_max_count > 1)
		{
			Log.Verbose("Read %i got %i Locations with max score %.2f like Location <%i %i>",
					read->ReadId, score_max_count, score_max,
					read->Scores[score_max_loc].Location.m_Location,
					read->Scores[score_max_loc].Location.m_RefId);

			read->EqualScoringCount = score_max_count;

			int esid = 1;
			for (int j = score_max_loc+1; j < read->nScores(); ++j)
			{
				if (read->Scores[j]->Score.f == score_max)
				{
					Log.Verbose("Read %i got max Score %.2f at Location <%i %i>", read->ReadId, score_max, read->Scores[j].Location.m_Location, read->Scores[j].Location.m_RefId);
					MappedRead * ntopread = new MappedRead(read->ReadId);
					ntopread->EqualScoringID = esid++;
					ntopread->EqualScoringCount = score_max_count;
					ntopread->Calculated = 1;
					ntopread->TopScore = 0;
					ntopread->AddScore(*read->Scores[j])->Read = ntopread;
					//NGM.bSWO.Write(&ntopread, 1);
					m_ReadBuffer.push_back(ntopread);
				}
			}
		}
	}
}


bool SW::CheckPairs(LocationScore * ls1, LocationScore * ls2, float & topScore, int & dst)
{
	if (ls1->Location.m_RefId == (ls2->Location.m_RefId ^ 0x1))
	{
		int distance = SequenceProvider.GetRefLen(ls1->Location.m_RefId) - 
			(ls1->Location.m_Location + ls2->Location.m_Location + 2*SequenceProvider.GetReadMaxLen());

		if (distance > m_PairMinDistance && distance < m_PairMaxDistance)
		{
			float f = ls1->Score.f + ls2->Score.f;
			if ( f > topScore )
			{
				topScore = f;
				dst = distance;
				return true;
			}
		}
	}
	return false;
}

void SW::PairedReadSelection(MappedRead * read1, MappedRead * read2)
{
	//Log.Green("Paired Read selection -> read %i: %i candidates / read %i: %i candidates", read1->ReadId, read1->nScores(), read2->ReadId, read2->nScores());

	/*for (int i = 0; i < read1->nScores(); ++i)
		Log.Message("Read %i c%i -> Location <%i, %i>, Score %f", read1->ReadId, i,
			read1->Scores[i]->Location.m_Location, read1->Scores[i]->Location.m_RefId,
			read1->Scores[i]->Score.f);

	for (int i = 0; i < read2->nScores(); ++i)
		Log.Message("Read %i c%i -> Location <%i, %i>, Score %f", read2->ReadId, i,
			read2->Scores[i]->Location.m_Location, read2->Scores[i]->Location.m_RefId,
			read2->Scores[i]->Score.f);*/

	float topScore = 0;
	int distance = 0;
	for (int i = 0; i < read1->nScores(); ++i)
	{
		for (int j = 0; j < read2->nScores(); ++j)
		{
			if (CheckPairs(read1->Scores[i], read2->Scores[j], topScore, distance))
			{
				read1->TopScore = i;
				read2->TopScore = j;
			}
		}
	}

	if (read1->TopScore != -1)
	{
		Log.Verbose("Read pairing found: R %i (#%i) at <%i, %i> Score %f, R %i (#%i) at <%i, %i> Score %f / Distance %i", 
			read1->ReadId, read1->TopScore, read1->Scores[read1->TopScore].Location.m_Location,
			read1->Scores[read1->TopScore].Location.m_RefId, read1->Scores[read1->TopScore].Score.f,
			read2->ReadId, read2->TopScore, read2->Scores[read2->TopScore].Location.m_Location,
			read2->Scores[read2->TopScore].Location.m_RefId, read2->Scores[read2->TopScore].Score.f,
			distance);

		//NGM.bSWO.Write(&read1, 1);
		//NGM.bSWO.Write(&read2, 1);
		m_ReadBuffer.push_back(read1);
		m_ReadBuffer.push_back(read2);
	}
	else
	{
		NGM.AddUnmappedRead(read1->ReadId, MFAIL_PAIRING);
		NGM.AddUnmappedRead(read2->ReadId, MFAIL_PAIRING);
		delete read1;
		delete read2;
	}
}

bool SW::CommitReads()
{
	int written = 0;
	//Log.Message("SW Thread %i committing finished reads", m_TID);
	while (!m_ReadBuffer.empty())
	{
		MappedRead * read = m_ReadBuffer.front();
		NGM.bSWO.Write(&read, 1);
		m_ReadBuffer.pop_front();
		++written;
	}
	Log.Message("SW Thread %i finished commit (%i Reads written)", m_TID, written);
	return true;
}
