#include "SequenceProvider.h"
#include "NGM.h"

#include <vector>
#include <map>
#include <string>

#include <string.h>

#include "Debug.h"

#undef module_name
#define module_name "SEQPROV"

_SequenceProvider * _SequenceProvider::pInstance = 0;

_SequenceProvider & _SequenceProvider::Instance()
{
	if (pInstance == 0)
		pInstance = new _SequenceProvider();

	return *pInstance;
}

bool checkChar(char c)
{
	return (c == 'A') | (c == 'C') | (c == 'T') | (c == 'G');
}

// checked die sequence seq auf ACGT, liefert das erste auftreten eines von diesen
// buchstaben unterschiedlichen zeichens ...4-basen-worte (32bit) verwenden?
int checkSequence(char * seq, uint len)
{
	for (uint n = 0; n < len; ++n)
	{
		if (!checkChar(*(seq+n)))
			return n;
		++n;
	}
	return -1;
}

void _SequenceProvider::PrepareInputFile(std::string filename)
{
	/*if (Config.GetInt("use_cache") == 1 && 
		FileExists(std::string(filename + std::string(".bs")).c_str())
		)
    return;*/

  //Launch("convert.exe", filename.c_str());
}

std::string CheckFile(std::string filename, char const * const name)
{
	if (!FileExists(filename.c_str()))
	{
		Log.Error("%s file not found (%s)", name, filename.c_str());
		Fatal();
	}
	return filename;
}

void _SequenceProvider::Init()
{
	if (!Config.Exists("ref"))
	{
		Log.Error("No reference file specified.");
		Fatal();
	}
	if (!Config.Exists("qry"))
	{
		Log.Error("No query file specified.");
		Fatal();
	}
	if (NGM.Paired() && !Config.Exists("pqry"))
	{
		Log.Error("No paired query file specified.");
		Fatal();
	}

	PrepareInputFile(refFileName);

	refIdxName = CheckFile(CheckFile(refBaseFileName = Config.GetString("ref"), "RefBase") + std::string(".idx"), "RefIdx");
	qryIdxName = CheckFile(CheckFile(qryBaseFileName = Config.GetString("qry"), "QryBase") + std::string(".idx"), "QryIdx");

	refIdxMap = CreateMapping(refIdxName.c_str(), pRefIdxData);
	qryIdxMap = CreateMapping(qryIdxName.c_str(), pQryIdxData);

	refHeader = (TIdxHeader*)pRefIdxData;
	refIdx = (TRefIdx*)(pRefIdxData+refHeader->HeaderLength);
	qryHeader = (TIdxHeader*)pQryIdxData;
	qryIdx = (TQryIdx*)(pQryIdxData+qryHeader->HeaderLength);

	if (refHeader->Signature != 0x524D474E)
	{
		Log.Error("Invalid reference file index");
		Fatal();
	}
	if (!(refHeader->VersionMajor == 1 && refHeader->VersionMinor == 1))
	{
		Log.Error("Invalid reference file version. Please rebuild input files.");
		Fatal();
	}
	if (qryHeader->Signature != 0x514D474E)
	{
		Log.Error("Invalid read file index");
		Fatal();
	}
	if (!(qryHeader->VersionMajor == 1 && qryHeader->VersionMinor == 1))
	{
		Log.Error("Invalid read file version. Please rebuild input files.");
		Fatal();
	}

	refFileName = CheckFile(Config.GetString("ref") + std::string(".bs"), "Reference");
	qryFileName = CheckFile(Config.GetString("qry") + std::string(".bs"), "Query");

	qryMaxLen = qryHeader->SeqFieldWidth;
	Log.Message("Reads: %i with len %i", qryHeader->SeqCount, qryMaxLen);

	refMap = CreateMapping(refFileName.c_str(), pRefData);
	qryMap = CreateMapping(qryFileName.c_str(), pQryData);

	refBaseMap = CreateMapping(refBaseFileName.c_str(), pRefBaseData );
	qryBaseMap = CreateMapping(qryBaseFileName.c_str(), pQryBaseData );

	refFileLen = FileSize(refFileName.c_str());
	qryFileLen = FileSize(qryFileName.c_str());

	refBaseFileLen = FileSize(refBaseFileName.c_str());
	qryBaseFileLen = FileSize(qryBaseFileName.c_str());

	if (NGM.DualStrand())
	{
		refCplFileName = CheckFile(Config.GetString("ref") + std::string(".bsc"), "Ref complement");

		refcplMap = CreateMapping(refCplFileName.c_str(), pRefCplData);
		refCplFileLen = FileSize(refCplFileName.c_str());

		if (refCplFileLen != refFileLen)
		{
			Log.Error("Different reference and reference complement file sizes (%i, %i)", refFileLen, refCplFileLen);
			Fatal();
		}
	}

	if (NGM.Paired())
	{
		prdBaseFileName = CheckFile(Config.GetString("pqry"), "Paired read base");
		prdFileName = CheckFile(prdBaseFileName + std::string(".bs"), "Paired read");
		prdIdxName = CheckFile(prdBaseFileName + ".idx", "Paired read index");

		prdFileLen = FileSize(prdFileName.c_str());
		prdBaseFileLen = FileSize(prdBaseFileName.c_str());

		prdIdxMap = CreateMapping(prdIdxName.c_str(), pPrdIdxData);
		prdMap = CreateMapping(prdFileName.c_str(), pPrdData );
		prdBaseMap = CreateMapping(prdBaseFileName.c_str(), pPrdBaseData);

		prdHeader = (TIdxHeader*)pPrdIdxData;
		prdIdx = (TQryIdx*)(pPrdIdxData+prdHeader->HeaderLength);
	}

	refCount = refHeader->SeqCount;
	if (NGM.DualStrand())
		refCount *= 2;

	qryCount = qryHeader->SeqCount;

	if ((qryFileLen != ((long)qryMaxLen * qryCount)) ||
		(NGM.Paired() && (qryFileLen != prdFileLen)) )
	{
		Log.Error("Wrong query file size.");
		Fatal();
	}

	if (NGM.Paired())
		qryCount *= 2;

	NGM.UpdateReadCount(qryCount);
}

char const * const _SequenceProvider::GetRefSequence(int n) const
{
	if (CheckRefNr(n))
	{
		char const * pBase = pRefData;
		if (NGM.DualStrand())
		{
			if (n & 1)
				pBase = pRefCplData;

			n >>= 1;
		}
		return pBase+refIdx[n].SeqStart;
	}
	else
		return 0;
}

char const * const _SequenceProvider::GetQrySequence(int n) const
{
	if (CheckQryNr(n))
	{
		char const * pBase = pQryData;
		if (NGM.Paired() && ( n & 1))
		{
			pBase = pPrdData;
			n >>= 1;
		}

		return pBase + (ulong)qryMaxLen * (ulong)n;
	}
	else
		return 0;
}

char const * _SequenceProvider::GetRefName(int n, int & len) const
{
	if (CheckRefNr(n))
	{
		if (NGM.DualStrand())
			n >>= 1;
		len = refIdx[n].DescLen;
		return pRefBaseData + refIdx[n].DescStart;
	}
	return 0;
}
char const * _SequenceProvider::GetQryName(int n, int & len) const
{
	if (CheckQryNr(n))
	{
		char const * pBase = pQryBaseData;

		ulong start = qryIdx[n].DescStart;
		len = qryIdx[n].DescLen;

		if (NGM.Paired() && ( n & 1))
		{
			pBase = pPrdBaseData;
			n >>= 1;

			start = prdIdx[n].DescStart;
			len = prdIdx[n].DescLen;
		}
		return pBase + start;
	}
	return 0;
}
char const * _SequenceProvider::GetQualityString(int n, int & len) const
{
	if (CheckQryNr(n))
	{
		char const * pBase = pQryBaseData;

		ulong start = qryIdx[n].QualityStart;
		len = qryIdx[n].QualityLen;

		if (NGM.Paired() && ( n & 1))
		{
			pBase = pPrdBaseData;
			n >>= 1;

			start = prdIdx[n].QualityStart;
			len = prdIdx[n].QualityLen;
		}

		return pBase + start;
	}
	return 0;
}

int _SequenceProvider::GetRefLen(int n) const
{
	if (GetRefSequence(n) == 0)
		return 0;
	if (NGM.DualStrand())
		n >>= 1;
	return refIdx[n].SeqLen;
}

int _SequenceProvider::GetQryLen(int n) const
{
	if (CheckQryNr(n))
		return strlen(GetQrySequence(n));
	else
		return 0;
}

int _SequenceProvider::GetReadMaxLen() const
{
	return qryMaxLen;
}

int _SequenceProvider::GetRefCount() const
{
	return refCount;
}

int _SequenceProvider::GetQryCount() const
{
	return qryCount;
}

bool _SequenceProvider::CheckQryNr(int n) const
{
	if ((n < 0) || (n > qryCount))
	{
		Log.Error("Tried to access invalid read (%i %x)", n, n);
		Fatal();
		return false;
	}
	return true;
}
bool _SequenceProvider::CheckRefNr(int n) const
{
	if (n >= refCount || n < 0)
	{
		Log.Error("Tried to access invalid reference sequence (%i %x).", n, n);
		Fatal();
		return false;
	}
	return true;
}

_SequenceProvider::_SequenceProvider()
{
	qryCount = -1;
}

_SequenceProvider::~_SequenceProvider()
{
	CloseMapping(qryIdxMap);
	CloseMapping(refIdxMap);

	CloseMapping(qryMap);
	CloseMapping(refMap);

	if (NGM.DualStrand())
		CloseMapping(refcplMap);

	if (NGM.Paired())
	{
		CloseMapping(prdIdxMap);
		CloseMapping(prdMap);
		CloseMapping(prdBaseMap);
	}
}

void _SequenceProvider::Cleanup()
{
	delete pInstance;
}

bool _SequenceProvider::IsValid(int n) const
{
	CheckQryNr(n);

	if (NGM.Paired() && ( n & 1 ))
	{
		return prdIdx[n].Flags & 0x1;
	}
	return qryIdx[n].Flags & 0x1;
}

bool _SequenceProvider::HasPair(int n) const
{
	return NGM.Paired() && IsValid( n ^ 0x1 );
}
